/*
**   IRC - Internet Relay Chat, source/ircdcore.c
**   Copyright (C) 1999 Chris Behrens
**
**   Portions Copyright (C) 1990 Jarkko Oikarinen and
**                      University of Oulu, Computing Center
**
**   This program is free software; you can redistribute it and/or modify
**   it under the terms of the GNU General Public License as published by
**   the Free Software Foundation; either version 1, or (at your option)
**   any later version.
**
**   This program is distributed in the hope that it will be useful,
**   but WITHOUT ANY WARRANTY; without even the implied warranty of
**   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**   GNU General Public License for more details.
**
**   You should have received a copy of the GNU General Public License
**   along with this program; if not, write to the Free Software
**   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "setup.h"

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <ctype.h>
#include <sys/types.h>
#include <signal.h>
#include <syslog.h>
#include <time.h>
#include <sys/time.h> 
#include <sys/resource.h>
#include <dlfcn.h>
#include <errno.h>
#include "config.h"

#define MAXFDLISTS	100

typedef struct _ircd_symbol
{
	char	*name;
#define IRCD_SYMBOL_FLAGS_NOTREQUIRED	0x001
	int		flags;
	void	*saddr;
} IRCD_SYMBOL;

static IRCD_SYMBOL Symbols[] =
{
#define IRCD_SYMBOL_MAIN_LOOP				0
	{	"_ircd_main_loop",			0,		NULL },
#define IRCD_SYMBOL_IRCDSO_VERSIONSTRING	1
	{	"_ircdso_versionstring",	0,		NULL },
#define IRCD_SYMBOL_IRCDSO_VERSION			2
	{	"_ircdso_version",			0,		NULL },
	{	NULL,						0,		NULL }
};

static void	*Data = NULL;
static int	Data_size = 0;
static int	Data_format_version = 0;
static char	*Ircdcore_versionstring = "1.0a1";
static int	Ircdcore_version = 1;

typedef struct _data
{
	int		Maxconnections;
	int		Maxfdlists;
} IRCD_DATA;

int main(int argc, char **argv)
{
#ifdef RLIMIT_NOFILE
	struct rlimit	r;
#endif
	int				firsttime = 1;
	void			*vptr;
	IRCD_SYMBOL		*sptr;
	int				i;
	int				ret;

	if (argc < 2)
	{
		fprintf(stderr, "Usage: ircd ircd.conf\n");
		exit(-1);
	}
	
#ifdef RLIMIT_NOFILE
	r.rlim_cur = MAXCONNECTIONS;
	r.rlim_max = MAXCONNECTIONS;
	setrlimit(RLIMIT_NOFILE, &r);
#endif

	{ /* I know this is lame, but I don't care. */
		int fds[MAXCONNECTIONS];

		for(i=0;i<MAXCONNECTIONS;i++)
		{
			ret = open("/dev/null", O_RDONLY);
			if (ret < 0)
				break;
			fds[i] = ret;
		}
		i--;
		if ((MAXCONNECTIONS < 1) || (i < 0) || (fds[i] < (MAXCONNECTIONS-1)))
		{
			int num = i < 0 ? 0 : fds[i]+1;

			fprintf(stderr,
			"MAXCONNECTIONS is set incorrectly.  I can only open %i file%s\n",
				num, num == 1 ? "" : "s");
			exit(-1);
		}
		while(i >= 0)
		{
			close(fds[i]);
			i--;
		}
	}

	openlog("ircd-comstud", LOG_PID|LOG_NDELAY, LOG_DAEMON);

	signal(SIGPIPE, SIG_IGN);
	signal(SIGHUP, SIG_IGN);

	Data = (char *)malloc(sizeof(IRCD_DATA));
	Data_size = sizeof(IRCD_DATA);
	((IRCD_DATA *)Data)->Maxconnections = MAXCONNECTIONS;
	((IRCD_DATA *)Data)->Maxfdlists = MAXFDLISTS;

	for(;;)
	{
		vptr = dlopen("./ircd.so", RTLD_NOW);
		if (!vptr)
		{
			const char *err = dlerror();

			syslog(LOG_ERR, "can't open ircd.so: %s", err);
			if (firsttime)
			{
				fprintf(stderr, "can't open ircd.so: %s\n", err);
				exit(-1);
			}
		}

		for(sptr=Symbols;sptr->name;sptr++)
		{
			sptr->saddr = dlsym(vptr, sptr->name);
			if (!sptr->saddr && !(sptr->flags & IRCD_SYMBOL_FLAGS_NOTREQUIRED))
			{
				fprintf(stderr,
					"required symbol \"%s\" is missing from ircd.so...\n",
					sptr->name);
				syslog(LOG_ERR,
					"required symbol \"%s\" is missing from ircd.so...",
					sptr->name);
				break;
			}
		}

		if (sptr->name)
		{
			dlclose(vptr);
			if (firsttime)
				exit(-1);
			sleep(1);
			continue;
		}

		if ((ret=*((int *)Symbols[IRCD_SYMBOL_IRCDSO_VERSION].saddr)) > Ircdcore_version)
		{
			syslog(LOG_ERR, "ircdso arg version %d > ircdcore arg version %d",
					ret, Ircdcore_version);
			fprintf(stderr, "ircdso arg version %d > ircdcore arg version %d\n",
					ret, Ircdcore_version);
			if (firsttime)
				exit(-1);
		}

		syslog(LOG_NOTICE, "version %s of ircd.so has been loaded",
					(char *)Symbols[IRCD_SYMBOL_IRCDSO_VERSIONSTRING].saddr);

		ret = ((int(*)())Symbols[IRCD_SYMBOL_MAIN_LOOP].saddr)(
					argc,
					argv,
					firsttime,
					Ircdcore_versionstring,
					&Data_format_version,
					&Data,
					&Data_size);

		dlclose(vptr);
		if (ret < 0)
		{
			sleep(1);
			continue;
		}
		firsttime = 0;
	}

	/* NOT REACHED */
	exit(0);
}
